/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::MMMPlayer
 * @author     Mirko Waechter ( mirko dot waechter at kit dot edu )
 * @date       2014
 * @copyright  http://www.gnu.org/licenses/gpl.txt
 *             GNU General Public License
 */

#include "MMMPlayer.h"
#include <MMM/Motion/MotionReaderXML.h>

#include <Core/core/exceptions/local/ExpressionException.h>

#include <Core/core/system/ArmarXDataPath.h>

#include <Core/observers/variant/Variant.h>
#include <Core/interface/observers/VariantBase.h>


using namespace armarx;
using namespace MMM;


void MMMPlayer::onInitComponent()
{
    usingProxy(getProperty<std::string>("KinematicUnitName").getValue());
    usingProxy("DebugObserver");

    motionFPS = 1;
    desiredFPS = 1;
    usePIDController = getProperty<bool>("UsePIDController").getValue();
    if(usePIDController)
        usingTopic(getProperty<std::string>("RobotNodeSetName").getValue());
}


void MMMPlayer::onConnectComponent()
{
    {
        ScopedLock lock(playerMutex);
        currentFrame = 0;
        MotionReaderXML reader;
        auto motionPath = getProperty<std::string>("MMMFile").getValue();
        ArmarXDataPath::getAbsolutePath(motionPath, motionPath);
        motion = reader.loadMotion(motionPath);
        if(!motion)
        {
            terminate();
            return;
        }
        jointNames = motion->getJointNames();
        if(motion->getNumFrames() > 1)
        {
            float frameDuration = motion->getMotionFrame(1)->timestep - motion->getMotionFrame(0)->timestep;
            ARMARX_CHECK_EXPRESSION(frameDuration != 0);
            ARMARX_CHECK_EXPRESSION(frameDuration > 0)
                    motionFPS = 1.0f/frameDuration;
            ARMARX_VERBOSE << VAROUT(motionFPS);
        }

        kin = getProxy<KinematicUnitInterfacePrx>(getProperty<std::string>("KinematicUnitName").getValue());
        debugObserver = getProxy<DebugObserverInterfacePrx>("DebugObserver");
        desiredFPS = getProperty<float>("FPS").getValue();

        start = 0.3;
        end = 1.26;
        direction = 1;

        jointOffets["Right Arm Shoulder2"] = 20.0/180.0*M_PI;
        jointOffets["Left Arm Shoulder2"] = 20.0/180.0*M_PI;

        NameControlModeMap modes;
        for (size_t i = 0; i < jointNames.size(); ++i) {
            const auto& jointName = jointNames.at(i);
            if(usePIDController)
            {
                modes[jointName] = eVelocityControl;
            }
            else
            {
                modes[jointName] = ePositionControl;
            }
        }
        kin->switchControlMode(modes);
    }

    maxVel = getProperty<float>("absMaximumVelocity").getValue();
    task = new PeriodicTask<MMMPlayer>(this, &MMMPlayer::run, 10, false, "MMMPlayerTask");
    motionStartTime = IceUtil::Time::now();
    task->start();
}


void MMMPlayer::run()
{
    ScopedLock lock(playerMutex);
    NameValueMap targetVelocities;



    currentMotionTime = (IceUtil::Time::now() - motionStartTime).toSecondsDouble() * 100 * desiredFPS / motionFPS;
    if(direction>0)
        currentFrame = currentMotionTime;
    else
        currentFrame = motion->getNumFrames() - (int)currentMotionTime;
    if(currentFrame >= (int)motion->getNumFrames() || currentFrame < 0)
    {
        if(getProperty<bool>("LoopPlayback").getValue())
        {
            direction *= -1;
            motionStartTime = IceUtil::Time::now();
            return;
        }
        else if(task)
        {
            return;
        }
    }
    double interpolationValue = currentMotionTime - (int)currentMotionTime;
    auto frame = motion->getMotionFrame(currentFrame);
    MMM::MotionFramePtr nextFrame;
    if(currentFrame +1 < (int)motion->getNumFrames())
        nextFrame = motion->getMotionFrame(currentFrame+1);
    targetPositionValues.clear();
    ARMARX_INFO << deactivateSpam(1) << "frame: " << currentFrame << " " << VAROUT(interpolationValue);
    StringVariantBaseMap debugTargetValues;
    ARMARX_CHECK_EXPRESSION((int)jointNames.size() == frame->joint.rows());
    for (size_t i = 0; i < jointNames.size(); ++i) {
        const auto& jointName = jointNames.at(i);
        auto& targetPosValue = targetPositionValues[jointName] = frame->joint[i];
        if(nextFrame)
            targetPosValue += interpolationValue * (nextFrame->joint[i] - frame->joint[i]);
        auto it = jointOffets.find(jointName);
        if(it != jointOffets.end())
            targetPosValue += it->second;
        assert(targetPosValue == targetPositionValues[jointName]);
        debugTargetValues[jointName] = new Variant(targetPosValue);
        float& targetVel = targetVelocities[jointName] =  0.0;
        auto vel = frame->joint_vel[i];
        if(vel != 0.0f)
        {
            targetVel = vel;
        }
        else if(nextFrame)
        {
            if(nextFrame->timestep - frame->timestep == 0)
                ARMARX_INFO << deactivateSpam() << "timestep delta is zero at frame " << currentFrame << " - cannot calculate velocity";
            targetVel = targetVelocities[jointName] = (nextFrame->joint[i] - frame->joint[i]) / (nextFrame->timestep - frame->timestep);


        }


        MMM::MotionFramePtr nextNextFrame;
        if(currentFrame + 2 < (int)motion->getNumFrames())
            nextNextFrame = motion->getMotionFrame(currentFrame+2);
        if(nextNextFrame)
        {
            float nextTargetVel = (nextNextFrame->joint[i] - nextFrame->joint[i]) / (nextNextFrame->timestep - nextFrame->timestep);
            targetVel += interpolationValue * (nextTargetVel - targetVel);
        }
        targetVel *= desiredFPS/motionFPS*direction;


//        targetVelocities[jointName] =  0.0;
        auto pid = PIDs.find(jointName);
        if(usePIDController && pid != PIDs.end())
        {
            auto cv = pid->second->getControlValue();
            targetVel += cv;
            if(cv > 20)
                ARMARX_INFO << "" << (jointName) << ": v:" << (targetVel) << " pos: " << targetPosValue;

//            debugObserver->setDebugDatafield(jointName, "p", new Variant(pid->second->previousError * pid->second->Kp));
//            debugObserver->setDebugDatafield(jointName, "i", new Variant(pid->second->integral* pid->second->Ki));
//            debugObserver->setDebugDatafield(jointName, "d", new Variant(pid->second->derivative * pid->second->Kd));
//            debugObserver->setDebugDatafield(jointName, "velocity", new Variant(targetVelocities[jointName]));
        }
        targetVel  = std::min<double>(maxVel/180.0*M_PI,targetVel);
        targetVel  = std::max<double>(-1 * maxVel/180.0*M_PI,targetVel);
    }
//    ARMARX_INFO << "target: " << targetPositionValues ;
    std::stringstream out;
    for(auto v : debugTargetValues)
        out << v.first << ": " << v.second->getFloat() << " " << targetPositionValues[v.first] << "\n" ;
//    ARMARX_INFO << "debugtarget: " << out.str() ;
    debugObserver->setDebugChannel("targetJointAngles", debugTargetValues);
    kin->setJointVelocities(targetVelocities);
    if(!usePIDController)
        kin->setJointAngles(targetPositionValues);
    ARMARX_VERBOSE << deactivateSpam(3) << "Playing frame " << currentFrame;
//    currentFrame+=direction;
}

void MMMPlayer::reportJointAngles(const NameValueMap &angles, bool, const Ice::Current &)
{
    ScopedLock lock(playerMutex);
    std::stringstream out;
    for(auto v : angles)
        out << v.first << ": " << v.second << " " << targetPositionValues[v.first] << "\n" ;
//    ARMARX_INFO  << deactivateSpam(0.5) << "jointangles reported: " << out.str();
    if(!usePIDController)
    {
        ARMARX_INFO << deactivateSpam() << "jointangles reporting DISABLED";
        return;

    }

    for(const auto &joint: angles)
    {

        const std::string& name = joint.first;
        auto it = PIDs.find(name);
        if(it == PIDs.end())
        {
            PIDs[name] = PIDControllerPtr(new PIDController(getProperty<float>("Kp").getValue(),
                                                                   getProperty<float>("Ki").getValue(),
                                                                   getProperty<float>("Kd").getValue()));
            it = PIDs.find(name);
        }
        ARMARX_INFO  << deactivateSpam() << name;
        PIDControllerPtr pid = it->second;
        pid->update(joint.second, targetPositionValues[name]);
    }
}

void MMMPlayer::onDisconnectComponent()
{
    if(task)
        task->stop();
    kin->stop();
}


void MMMPlayer::onExitComponent()
{

}

PropertyDefinitionsPtr MMMPlayer::createPropertyDefinitions()
{
    return PropertyDefinitionsPtr(new MMMPlayerPropertyDefinitions(
                                      getConfigIdentifier()));
}
