/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2 of
* the License, or (at your option) any later version.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @package    ArmarX::
* @author     Philipp Schmidt
* @date       2015
* @license    http://www.gnu.org/licenses/gpl.txt
*             GNU General Public License

*/

#include "RobotIKGuiPlugin.h"
#include "RobotViewer.h"

/* ArmarX includes */
#include <Core/core/system/ArmarXDataPath.h>
#include <Core/core/ArmarXObjectScheduler.h>
#include <Core/core/system/cmake/CMakePackageFinder.h>
#include <Core/core/application/Application.h>

/* Virtual Robot includes */
#include <VirtualRobot/XML/RobotIO.h>
#include <VirtualRobot/Visualization/CoinVisualization/CoinVisualizationFactory.h>

/* Coin includes */
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/SbViewportRegion.h>

#define ROBOT_UPDATE_TIMER_MS 33
#define TEXTFIELD_UPDATE_TIMER_MS 200

armarx::RobotIKGuiPlugin::RobotIKGuiPlugin()
{
    addWidget<RobotIKWidgetController>();
}

armarx::RobotIKWidgetController::RobotIKWidgetController() : startUpCameraPositioning(true)
{
    //Initialize Gui
    ui.setupUi(getWidget());
    getWidget()->setEnabled(false);

    //Alignment for ui
    this->ui.verticalLayout->setAlignment(Qt::AlignTop);

    //Label color
    this->ui.reachableLabel->setStyleSheet("QLabel { color : red; }");


    //Component names
    robotStateComponentName = "";
    robotIKComponentName = "";
    kinematicUnitComponentName = "";

    manipSeparator = NULL;
    currentSeparator = NULL;

    tcpManip = NULL;
    tcpCurrent = NULL;
    tcpCurrentTransform = NULL;
    tcpDesired = NULL;
    tcpManipTransform = NULL;
}

void armarx::RobotIKWidgetController::onInitComponent()
{
    //Prepare proxies
    usingProxy(robotStateComponentName);
    usingProxy(robotIKComponentName);
    usingProxy(kinematicUnitComponentName);
}

void armarx::RobotIKWidgetController::onConnectComponent()
{
    robotStateComponentPrx = getProxy<RobotStateComponentInterfacePrx>(robotStateComponentName);
    kinematicUnitInterfacePrx = getProxy<KinematicUnitInterfacePrx>(kinematicUnitComponentName);
    robotIKPrx = getProxy<RobotIKInterfacePrx>(robotIKComponentName);

    //Load robot
    robot = loadRobot(getIncludePaths());
    if(!robot)
    {
        getObjectScheduler()->terminate();
        if(getWidget()->parentWidget())
        {
            getWidget()->parentWidget()->close();
        }
        ARMARX_ERROR << "RobotIKPlugin: Unable to load robot file! Terminating." << std::endl;
        return;
    }

    //Get visualization for our robot and add it to scene graph
    CoinVisualizationPtr robotViewerVisualization = robot->getVisualization<VirtualRobot::CoinVisualization>();
    this->ui.robotViewer->getRobotViewer()->getRootNode()->addChild(robotViewerVisualization->getCoinVisualization());

    //Make a timer update robo pose every time tick
    SoSensorManager* sensor_mgr = SoDB::getSensorManager();
    robotUpdateSensor = new SoTimerSensor(robotUpdateTimerCB, this);
    robotUpdateSensor->setInterval(SbTime(ROBOT_UPDATE_TIMER_MS / 1000.0f));
    sensor_mgr->insertTimerSensor(robotUpdateSensor);

    //And a timer to update the labels
    textFieldUpdateSensor = new SoTimerSensor(textFieldUpdateTimerCB, this);
    textFieldUpdateSensor->setInterval(SbTime(TEXTFIELD_UPDATE_TIMER_MS / 1000.0f));
    sensor_mgr->insertTimerSensor(textFieldUpdateSensor);

    //Get all kinematic chain descriptions and add them to the combo box
    auto robotNodeSets = robot->getRobotNodeSets();
    for(VirtualRobot::RobotNodeSetPtr s : robotNodeSets) {
        this->ui.comboBox->addItem(QString::fromStdString(s->getName()));
    }

    connectSlots();
    enableMainWidgetAsync(true);
}

armarx::StringList armarx::RobotIKWidgetController::getIncludePaths()
{
    StringList includePaths;

    try {
        StringList packages = robotStateComponentPrx->getArmarXPackages();
        packages.push_back(Application::GetProjectName());
        for(const std::string &projectName : packages)
        {
            if(projectName.empty())
                continue;
            CMakePackageFinder project(projectName);
            StringList projectIncludePaths;
            auto pathsString = project.getDataDir();
            boost::split(projectIncludePaths,
                         pathsString,
                         boost::is_any_of(";,"),
                         boost::token_compress_on);
            includePaths.insert(includePaths.end(), projectIncludePaths.begin(), projectIncludePaths.end());
        }
    } catch (...)
    {
        ARMARX_ERROR << "Unable to retrieve robot filename." << std::endl;
    }

    return includePaths;
}

VirtualRobot::RobotPtr armarx::RobotIKWidgetController::loadRobot(StringList includePaths)
{
    try
    {
        std::string rfile = robotStateComponentPrx->getRobotFilename();
        ArmarXDataPath::getAbsolutePath(rfile, rfile, includePaths);
        return VirtualRobot::RobotIO::loadRobot(rfile);
    }
    catch(...)
    {
        ARMARX_ERROR << "Unable to load robot from file" << std::endl;
        return VirtualRobot::RobotPtr();
    }
}

void armarx::RobotIKWidgetController::onDisconnectComponent()
{
    robotStateComponentPrx = NULL;
    kinematicUnitInterfacePrx = NULL;
    robotIKPrx = NULL;
}

void armarx::RobotIKWidgetController::onExitComponent()
{
    enableMainWidgetAsync(false);
}

QPointer<QDialog> armarx::RobotIKWidgetController::getConfigDialog(QWidget *parent)
{
    if(!dialog)
    {
        dialog = new RobotIKConfigDialog(parent);
    }
    return qobject_cast<RobotIKConfigDialog*>(dialog);
}

void armarx::RobotIKWidgetController::loadSettings(QSettings *settings)
{

}

void armarx::RobotIKWidgetController::saveSettings(QSettings *settings)
{

}

void armarx::RobotIKWidgetController::configured()
{
    robotStateComponentName = dialog->robotStateComponentProxyFinder->getSelectedProxyName().trimmed().toStdString();
    robotIKComponentName = dialog->robotIKComponentProxyFinder->getSelectedProxyName().trimmed().toStdString();
    kinematicUnitComponentName = dialog->kinematicUnitComponentProxyFinder->getSelectedProxyName().trimmed().toStdString();
}

void armarx::RobotIKWidgetController::solveIK()
{
    auto targetJointAngles = getIKSolution();

    if(targetJointAngles.isReachable) {
        //Switch all control modes to ePositionControl
        std::vector<VirtualRobot::RobotNodePtr> rn = robot->getRobotNodeSet(this->ui.comboBox->currentText().toStdString())->getAllRobotNodes();
        NameControlModeMap jointModes;
        NameValueMap jointAngles;
        for (unsigned int i=0;i<rn.size();i++)
        {
            jointModes[rn[i]->getName()] = ePositionControl;
            jointAngles[rn[i]->getName()]=0.0f;
        }
        kinematicUnitInterfacePrx->switchControlMode(jointModes);
        kinematicUnitInterfacePrx->setJointAngles(targetJointAngles.jointAngles);
    }
}


void armarx::RobotIKWidgetController::connectSlots()
{
    connect(ui.resetManip, SIGNAL(clicked()), this, SLOT(resetManip()), Qt::QueuedConnection);
    connect(ui.pushButton, SIGNAL(clicked()), this, SLOT(solveIK()), Qt::QueuedConnection);
    connect(ui.comboBox, SIGNAL(currentIndexChanged(QString)), this, SLOT(kinematicChainChanged(QString)), Qt::UniqueConnection);
}

void armarx::RobotIKWidgetController::robotUpdateTimerCB(void *data, SoSensor *sensor)
{
    RobotIKWidgetController* controller = static_cast<RobotIKWidgetController*>(data);
    if (!controller)
        return;
    try
    {
        armarx::RemoteRobot::synchronizeLocalClone(controller->robot,controller->robotStateComponentPrx);

        if(controller->currentSeparator) {
            Eigen::Matrix4f tcpMatrix = controller->robot->getRobotNodeSet(controller->ui.comboBox->currentText().toStdString())->getTCP()->getGlobalPose();
            //Apply tcp position to indicator
            tcpMatrix(0, 3) /= 1000;
            tcpMatrix(1, 3) /= 1000;
            tcpMatrix(2, 3) /= 1000;
            controller->tcpCurrentTransform->setMatrix(VirtualRobot::CoinVisualizationFactory::getSbMatrix(tcpMatrix));
        }

        if(controller->startUpCameraPositioning) {
            controller->ui.robotViewer->getRobotViewer()->cameraViewAll();
            controller->startUpCameraPositioning = false;
        }
    } catch (...){};
}

void armarx::RobotIKWidgetController::textFieldUpdateTimerCB(void *data, SoSensor *sensor)
{
    RobotIKWidgetController* controller = static_cast<RobotIKWidgetController*>(data);
    if (!controller)
        return;

    if(controller->currentSeparator) {
        Eigen::Matrix4f tcpMatrix = controller->robot->getRobotNodeSet(controller->ui.comboBox->currentText().toStdString())->getTCP()->getGlobalPose();

        //Set text label to tcp matrix
        std::stringstream buffer;
        buffer << tcpMatrix;
        std::string matrixText = buffer.str();
        controller->ui.currentPoseMatrix->setText(QString::fromStdString(matrixText));

        //Set text label for desired tcp pose
        //Therefore calculate current manipulator pose
        SoGetMatrixAction* action = new SoGetMatrixAction(SbViewportRegion());
        SoSearchAction sa;
        sa.setNode(controller->tcpDesired);
        sa.setSearchingAll( TRUE );              // Search all nodes
        SoBaseKit::setSearchingChildren( TRUE ); // Even inside nodekits
        sa.apply(controller->ui.robotViewer->getRobotViewer()->getRootNode());

        action->apply(sa.getPath());

        SbMatrix matrix = action->getMatrix();

        Eigen::Matrix4f mat = Eigen::Matrix4f::Identity();
        mat (0,0) = matrix[0][0];
        mat (0,1) = matrix[1][0];
        mat (0,2) = matrix[2][0];
        mat (0,3) = matrix[3][0] * 1000;

        mat (1,0) = matrix[0][1];
        mat (1,1) = matrix[1][1];
        mat (1,2) = matrix[2][1];
        mat (1,3) = matrix[3][1] * 1000;

        mat (2,0) = matrix[0][2];
        mat (2,1) = matrix[1][2];
        mat (2,2) = matrix[2][2];
        mat (2,3) = matrix[3][2] * 1000;

        mat (3,0) = matrix[0][3];
        mat (3,1) = matrix[1][3];
        mat (3,2) = matrix[2][3];
        mat (3,3) = matrix[3][3];

        //Set text label to manip matrix
        std::stringstream buffer2;
        buffer2 << mat;
        std::string matrixText2 = buffer2.str();
        controller->ui.desiredPoseMatrix->setText(QString::fromStdString(matrixText2));

        delete(action);
    }
}

void armarx::RobotIKWidgetController::kinematicChainChanged(const QString &arg1)
{
    this->ui.moveTCP->setEnabled(true);
    this->ui.reachableLabel->setEnabled(true);
    this->ui.reachableLabel->setText(QString::fromStdString("Pose reachable!"));
    this->ui.reachableLabel->setStyleSheet("QLabel { color : green; }");
    this->ui.pushButton->setEnabled(true);
    this->ui.currentPose->setEnabled(true);
    this->ui.currentPoseMatrix->setEnabled(true);
    this->ui.desiredPose->setEnabled(true);
    this->ui.desiredPoseMatrix->setEnabled(true);
    this->ui.resetManip->setEnabled(true);


    Eigen::Matrix4f mat = robot->getRobotNodeSet(arg1.toStdString())->getTCP()->getGlobalPose();
    mat(0, 3) /= 1000;
    mat(1, 3) /= 1000;
    mat(2, 3) /= 1000;

    if(currentSeparator) {
        this->ui.robotViewer->getRobotViewer()->getRootNode()->removeChild(currentSeparator);
    }
    if(manipSeparator) {
        this->ui.robotViewer->getRobotViewer()->getRootNode()->removeChild(manipSeparator);
    }

    manipSeparator = new SoSeparator;
    currentSeparator = new SoSeparator;

    tcpCurrentColor = new SoMaterial;
    tcpCurrentColor->transparency = 0.6f;
    tcpCurrent = new SoSphere;
    tcpCurrent->radius = 0.06f;
    tcpDesiredColor = new SoMaterial;
    tcpDesiredColor->ambientColor.setValue(0, 1, 0);
    tcpDesiredColor->transparency = 0.3f;
    tcpDesired = new SoSphere;
    tcpDesired->radius = 0.06f;
    tcpManip = new SoTransformerManip;
    SoSeparator* nullSep = new SoSeparator;

    //Make all scale knobs disappear
    tcpManip->getDragger()->setPart("scale1", nullSep);
    tcpManip->getDragger()->setPart("scale2", nullSep);
    tcpManip->getDragger()->setPart("scale3", nullSep);
    tcpManip->getDragger()->setPart("scale4", nullSep);
    tcpManip->getDragger()->setPart("scale5", nullSep);
    tcpManip->getDragger()->setPart("scale6", nullSep);
    tcpManip->getDragger()->setPart("scale7", nullSep);
    tcpManip->getDragger()->setPart("scale8", nullSep);

    tcpManip->getDragger()->addFinishCallback(manipFinishCallback, this);
    tcpCurrentTransform = new SoTransform;
    tcpManipTransform = new SoTransform;

    manipSeparator->addChild(tcpManipTransform);
    manipSeparator->addChild(tcpManip);
    manipSeparator->addChild(tcpDesiredColor);
    manipSeparator->addChild(tcpDesired);

    currentSeparator->addChild(tcpCurrentTransform);
    currentSeparator->addChild(tcpCurrentColor);
    currentSeparator->addChild(tcpCurrent);

    this->ui.robotViewer->getRobotViewer()->getRootNode()->insertChild(currentSeparator, 0);
    this->ui.robotViewer->getRobotViewer()->getRootNode()->insertChild(manipSeparator, 0);

    /*tcpCurrentTransform->translation.setValue(mat(0, 3) / 1000, mat(1, 3) / 1000, mat(2, 3) / 1000);
    tcpCurrentTransform->rotation.setValue(SbRotation());*/

    VirtualRobot::CoinVisualizationFactory::getSbMatrix(mat);

    tcpCurrentTransform->setMatrix(VirtualRobot::CoinVisualizationFactory::getSbMatrix(mat));
    tcpManipTransform->setMatrix(VirtualRobot::CoinVisualizationFactory::getSbMatrix(mat));

    /*tcpManip->translation.setValue(mat(0, 3) / 1000, mat(1, 3) / 1000, mat(2, 3) / 1000);
    tcpManip->rotation.setValue(SbRotation(VirtualRobot::CoinVisualizationFactory::getSbMatrix(mat)));*/
}

void armarx::RobotIKWidgetController::resetManip()
{
    //Triggers reset of manipulator in kinematicChainChanged
   kinematicChainChanged(this->ui.comboBox->currentText());
}

void armarx::RobotIKWidgetController::manipFinishCallback(void* data, SoDragger* dragger) {
    RobotIKWidgetController* controller = static_cast<RobotIKWidgetController*>(data);

    auto targetJointAngles = controller->getIKSolution();

    if(targetJointAngles.isReachable) {
        //Green
        controller->tcpDesiredColor->ambientColor.setValue(0, 1, 0);
        controller->ui.reachableLabel->setText(QString::fromStdString("Pose reachable!"));
        controller->ui.reachableLabel->setStyleSheet("QLabel { color : green; }");
        controller->ui.pushButton->setEnabled(true);
    }
    else {
        //Red
        controller->tcpDesiredColor->ambientColor.setValue(1, 0, 0);
        controller->ui.reachableLabel->setText(QString::fromStdString("Pose unreachable!"));
        controller->ui.reachableLabel->setStyleSheet("QLabel { color : red; }");
        controller->ui.pushButton->setEnabled(false);
    }
}

armarx::ExtendedIKResult armarx::RobotIKWidgetController::getIKSolution()
{
    SoGetMatrixAction* action = new SoGetMatrixAction(SbViewportRegion());
    SoSearchAction sa;
    sa.setNode(tcpDesired);
    sa.setSearchingAll( TRUE );              // Search all nodes
    SoBaseKit::setSearchingChildren( TRUE ); // Even inside nodekits
    sa.apply(this->ui.robotViewer->getRobotViewer()->getRootNode());

    action->apply(sa.getPath());

    SbMatrix matrix = action->getMatrix();

    Eigen::Matrix4f mat = Eigen::Matrix4f::Identity();
    mat (0,0) = matrix[0][0];
    mat (0,1) = matrix[1][0];
    mat (0,2) = matrix[2][0];
    mat (0,3) = matrix[3][0] * 1000;

    mat (1,0) = matrix[0][1];
    mat (1,1) = matrix[1][1];
    mat (1,2) = matrix[2][1];
    mat (1,3) = matrix[3][1] * 1000;

    mat (2,0) = matrix[0][2];
    mat (2,1) = matrix[1][2];
    mat (2,2) = matrix[2][2];
    mat (2,3) = matrix[3][2] * 1000;

    mat (3,0) = matrix[0][3];
    mat (3,1) = matrix[1][3];
    mat (3,2) = matrix[2][3];
    mat (3,3) = matrix[3][3];

//    mat = robot->getRootNode()->toLocalCoordinateSystem(mat);
    return(robotIKPrx->computeExtendedIKGlobalPose(this->ui.comboBox->currentText().toStdString(),
                                                                              new armarx::Pose(mat), eAll));
}

Q_EXPORT_PLUGIN2(armarx_gui_RobotIKGuiPlugin, armarx::RobotIKGuiPlugin)
