/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarX::Core
 * @author     Christian Boege (boege at kit dot edu)
 * @copyright  2011 Christian Boege
 * @license    http://www.gnu.org/licenses/gpl.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_CORE_KINEMATICUNIT_SLICE_
#define _ARMARX_CORE_KINEMATICUNIT_SLICE_

#include <RobotAPI/interface/units/UnitInterface.ice>

#include <Core/interface/core/UserException.ice>
#include <Core/interface/core/BasicTypes.ice>

module armarx
{
    struct RangeViolation
    {
        float rangeFrom;
        float rangeTo;
        float actualValue;
    };

    sequence<RangeViolation> RangeViolationSequence;

    exception OutOfRangeException extends UserException
    {
        RangeViolationSequence violation;
    };

    enum ControlMode
    {
        eDisabled,
        eUnknown,
        ePositionControl,
        eVelocityControl,
        eTorqueControl,
        ePositionVelocityControl
    };

    enum OperationStatus
    {
        eOffline,
        eOnline,
        eInitialized
    };
    
    enum ErrorStatus
    {
        eOk,
        eWarning,
        eError
    };
    
    struct JointStatus
    {
        OperationStatus operation;
        ErrorStatus error;
        
        bool enabled;
        bool emergencyStop;
    };

    exception ControlModeNotSupportedException extends UserException
    {
        ControlMode mode;
    };

    exception KinematicUnitUnavailable extends ResourceUnavailableException
    {
        StringStringDictionary nodeOwners;
    };

    exception KinematicUnitNotOwnedException extends ResourceNotOwnedException
    {
        StringSequence nodes;
    };

    dictionary<string, float> NameValueMap;
    dictionary<string, ControlMode> NameControlModeMap;
    dictionary<string, JointStatus> NameStatusMap;

    interface KinematicUnitInterface extends SensorActorUnitInterface
    {
        /*
         * Requesting and releasing joints guarantees that a joint is controlled by exactly
         * one component at any time
         * 
         * The request only affects the active access, sensor values can always be read.
         * 
         * TODO: The actual access has to be checked using a token.
         */
        void requestJoints(StringSequence joints) throws KinematicUnitUnavailable;
        void releaseJoints(StringSequence joints) throws KinematicUnitNotOwnedException;
        
        void switchControlMode(NameControlModeMap targetJointModes) throws ControlModeNotSupportedException;
        
        /*
         * Depending on the chosen control mode, one or more of the following functions need to be called
         * to set all necessary parameters.
         * 
         * The motion starts immediately.
         */
        void setJointAngles(NameValueMap targetJointAngles) throws OutOfRangeException;
        void setJointVelocities(NameValueMap targetJointVelocities) throws OutOfRangeException;
        void setJointTorques(NameValueMap targetJointTorques) throws OutOfRangeException;
        void setJointAccelerations(NameValueMap targetJointAccelerations) throws OutOfRangeException;
        void setJointDecelerations(NameValueMap targetJointDecelerations) throws OutOfRangeException;
        
        /*
         * NYI
         */
        //void set Trajectory(...);
    };

    interface KinematicUnitListener
    {
        void reportControlModeChanged(NameControlModeMap actualJointModes, bool aValueChanged);

        void reportJointAngles(NameValueMap actualJointAngles, bool aValueChanged);
        void reportJointVelocities(NameValueMap actualJointVelocities, bool aValueChanged);
        void reportJointTorques(NameValueMap actualJointTorques, bool aValueChanged);
        void reportJointAccelerations(NameValueMap actualJointAccelerations, bool aValueChanged);
        void reportJointCurrents(NameValueMap actualJointCurrents, bool aValueChanged);
        void reportJointMotorTemperatures(NameValueMap actualJointMotorTemperatures, bool aValueChanged);
        void reportJointStatuses(NameStatusMap actualJointStatuses, bool aValueChanged);
    };
    
    /*interface KinematicUnitHandInterface extends KinematicUnitInterface
    {
        void open();
        void close();
        void resetTactileSensors(bool r);

        void getObjectPoseInHandFrame();
        void getHandPoseInObjectFrame();
    };*/

    /*interface KinematicUnitHandListener
    {
        void reportTactileSensorData(NameValueMap actualTactileSensorData);
        void reportTactileSensorRawData(NameValueMap actualTactileSensorRawData);
        void reportTactileSensorContactData(NameValueMap actualTactileSensorContactData);
    };*/

};

#endif
